/* libexs/list.c */

/*
 * Copyright 2012, 2013, 2014 by
 *	University of New Hampshire Interoperability Laboratory
 *	121 Technology Drive, Suite 2
 *	Durham, NH 03824-4716
 *	phone: 1-603-862-0090
 *	fax: 1-603-862-4181
 *	e-mail: info@iol.unh.edu
 * All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grant No. OCI-1127228.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

/* Implementation in this file based on the two-lock concurrent queue algorithm
 * proposed in
 *   [MS1996] M. Michael and M. Scott. Simple, Fast, and Practical Non-Blocking
 *   and Blocking Concurrent Queue Algorithms, 1996.
 *
 * Concurrent calls to exs_p_list_add_tail and exs_p_list_del_head may be made
 * on the same list without any external synchronization.
 *
 * When implementing list methods, if both the head and tail mutexes must be
 * locked, the head mutex must always be locked before the tail mutex (and
 * obviously unlocked in the reverse order).
 *
 * List invariants from [MS1996]:
 * 1. The linked list is always connected.
 * 2. Nodes are only inserted after the last node in the linked list.
 * 3. Nodes are only deleted from the beginning of the linked list.
 * 4. q->head always points to the first node in the linked list.
 * 5. q->tail always points to a node in the linked list.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include <limits.h>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <atomic.h>
#include <list.h>

/* Initializes the list. May fail with ENOMEM if not enough space is available
 * on the heap to initialize the list. On success, returns 0, else returns
 * -1. The only currently supported flag is OUR_LIST_BLOCKING, which creates a
 * list that will block if exs_p_list_del_head() is called when the list is
 * empty. If the flag is not provided, the exs_p_list_del_head() function will
 * return NULL if the list is empty. */
int
exs_p_list_init(struct our_list_head *list, int flags)
{
	if ((flags & ~OUR_LIST_ALL_FLAGS)) {
		errno = EINVAL;
		return -1;
	}

	list->empty_head = NULL;
	list->tail = list->head = calloc(1, sizeof(*list->head));
	if (!list->head) {
		goto cleanup0;
	}
	list->empty_tail = list->empty_head
		= calloc(1, sizeof(*list->empty_head));
	if (!list->empty_head) {
		goto cleanup0;
	}
	if ((errno = pthread_mutex_init(&list->head_mutex, NULL)) != 0) {
		goto cleanup0;
	}
	if ((errno = pthread_mutex_init(&list->tail_mutex, NULL)) != 0) {
		goto cleanup1;
	}
	if (flags & OUR_LIST_BLOCKING) {
		if ((errno = pthread_cond_init(&list->empty_cv, NULL)) != 0) {
			goto cleanup2;
		}
		list->wait_count = 0;
	}
	list->flags = flags;
	return 0;

cleanup2:
	pthread_mutex_destroy(&list->tail_mutex);

cleanup1:
	pthread_mutex_destroy(&list->head_mutex);

cleanup0:
	free(list->head);
	list->head = list->tail = NULL;
	free(list->empty_head);
	list->empty_head = list->empty_tail = NULL;
	return -1;
}

/* Destroys the list. The caller must ensure that no other operations are
 * occuring on the list through some external means. */
void
exs_p_list_destroy(struct our_list_head *list)
{
	struct our_list_slot *ptr;

	list->flags &= ~OUR_LIST_BLOCKING;
	pthread_cond_destroy(&list->empty_cv);
	pthread_mutex_destroy(&list->tail_mutex);
	pthread_mutex_destroy(&list->head_mutex);

	while (list->empty_head) {
		ptr = list->empty_head;
		list->empty_head = ptr->next;
		free(ptr);
	}
	while (list->head) {
		ptr = list->head;
		list->head = ptr->next;
		free(ptr);
	}
	memset(list, 0, sizeof(*list));
}

/* Gets a node off of the free list, or creates a new one if none are available.
 * list->tail_mutex MUST be locked by this thread when this function is
 * called. This function does not call any POSIX cancellation points. */
static inline struct our_list_slot *
get_free_node(struct our_list_head *list)
{
	struct our_list_slot *node = list->empty_head;
	if (!node->next) {
		return malloc(sizeof(struct our_list_slot));
	}
	list->empty_head = node->next;
	return node;
}

/* Adds the given value to the tail of the list.  Does not call any POSIX
 * cancellation points.  Returns nonnegative on success or the negative value of
 * errno on failure.  Returns 1 if the list was empty prior to adding this
 * element and 0 if the list was non-empty prior to adding this element.  Note
 * that 1 may be returned spuriously due to another threads removing elements
 * while this function is adding the new element, but a return value of 0
 * guarantees that the list was non-empty prior to adding this element.
 *
 * May fail with -ENOMEM if there are no nodes on the free list and
 * there is insufficient heap memory to create a new node. This function does
 * not call any POSIX cancellation points. */
int
exs_p_list_add_tail(struct our_list_head *list, void *value)
{
	struct our_list_slot *new_node;
	int ret;
	
	pthread_mutex_lock(&list->tail_mutex);
	new_node = get_free_node(list);
	if (new_node == NULL) {
		ret = errno;
		goto out;
	}
	new_node->value = value;
	new_node->next = NULL;

	list->tail->next = new_node;
	ret = (list->tail == list->head) ? OUR_LIST_WAS_EMPTY : 0;
	list->tail = new_node;

	if (list->flags & OUR_LIST_BLOCKING) {
		if (list->wait_count > 0) {
			pthread_cond_signal(&list->empty_cv);
		}
	}
	pthread_mutex_unlock(&list->tail_mutex);

out:
	return ret;
}

/* Puts the node on the free list. list->tail_mutex MUST be locked by the
 * calling thread before calling this function. This function will not call any
 * POSIX cancellation points. */
static inline void
put_free_node(struct our_list_head *list, struct our_list_slot *node)
{
	node->next = node->value = NULL;
	list->empty_tail->next = node;
	list->empty_tail = node;
}

/* Cancellation handler that is pushed only in block_until_not_empty().  Unlocks
 * both list mutexes before exiting. This function does not call any POSIX
 * cancellation points. */
static void
list_wait_cancel_handler(void *ptr)
{
	struct our_list_head *list = ptr;

	pthread_mutex_unlock(&list->tail_mutex);
	pthread_mutex_unlock(&list->head_mutex);
}

/* Returns 0 if and only if the list is not empty by the specified deadline,
 * returns -1 if we must exit early.  Head mutex MUST be locked, tail mutex
 * MUST NOT be locked. This function calls POSIX cancellation points and uses a
 * POSIX cancellation handler to release resources used by this function. */
static int
block_until_not_empty(struct our_list_head *list, struct timespec *deadline)
{
	int ret = 0;

	pthread_cleanup_push(&list_wait_cancel_handler, list);
	while ((list->flags & OUR_LIST_BLOCKING) && !list->head->next) {
		ret = our_mutex_timedlock(&list->tail_mutex, deadline);
		if (ret) {
			errno = ret;
			ret = -1;
			break;
		}

		if (!(list->flags & OUR_LIST_BLOCKING) || list->head->next) {
			ret = 0;
		} else {
			++list->wait_count;
#if 0
			fprintf(stderr, " WAIT list empty_cv head->next %p\n",
					list->head->next);
#endif
			ret = our_cv_timedwait(&list->empty_cv,
						&list->tail_mutex, deadline);
#if 0
			fprintf(stderr, "AWAKE list empty_cv head->next %p\n",
					list->head->next);
#endif
			--list->wait_count;
		}
		pthread_mutex_unlock(&list->tail_mutex);
		if (ret) {
			errno = ret;
			ret = -1;
			break;
		}
	}
	pthread_cleanup_pop(0);

	if (!(list->flags & OUR_LIST_BLOCKING)) {
		errno = EINVAL;
	}

	return ret;
}

/* Removes the first element from the list and returns it. Requires the
 * head_mutex to be locked. This function does not call any POSIX cancellation
 * points. */
static void *
del_head_common(struct our_list_head *list)
{
	struct our_list_slot *node, *new_head;
	void *value;

	node = list->head;
	new_head = node->next;
	if (!new_head) {
		return NULL;
	}

	value = new_head->value;
	list->head = new_head;
	put_free_node(list, node);
	return value;
}

/* Removes the first item from the list and returns its value, blocking if the
 * list is empty until an element is added to the list.  May not be called on a
 * non-blocking list.  This function may call POSIX cancellation points. */
void *
exs_p_list_del_head_or_wait(struct our_list_head *list,
		struct timespec *deadline)
{
	void *value;
	int blocking = (list->flags & OUR_LIST_BLOCKING);
	int ret;

	if (!blocking) {
		errno = EINVAL;
		return NULL;
	}

	ret = our_mutex_timedlock(&list->head_mutex, deadline);
	if (ret != 0) {
		errno = ret;
		return NULL;
	}

	if (!list->head->next) {
		if (block_until_not_empty(list, deadline) != 0) {
			value = NULL;
			goto end;
		}
	}
	value = del_head_common(list);

end:
	pthread_mutex_unlock(&list->head_mutex);
	return value;
}

/* Removes the first item from the list and returns its value.  Does not call
 * any POSIX cancellation points.  Returns NULL if the list is empty. */
void *
exs_p_list_del_head(struct our_list_head *list)
{
	void *value;
	int ret;

	ret = pthread_mutex_lock(&list->head_mutex);
	if (ret != 0) {
		errno = ret;
		return NULL;
	}

	value = del_head_common(list);

	pthread_mutex_unlock(&list->head_mutex);
	return value;
}

/* Returns the value of the first item from the list without removing it.  Does
 * not call any POSIX cancellation points.  Returns NULL if the list is
 * empty. */
void *
exs_p_list_peek_head(struct our_list_head *list)
{
	struct our_list_slot *node;
	void *value;
	int ret;

	ret = pthread_mutex_lock(&list->head_mutex);
	if (ret != 0) {
		return NULL;
	}

	node = list->head->next;
	value = node ? node->value : NULL;

	pthread_mutex_unlock(&list->head_mutex);
	return value;
}

/* Interrupts all threads waiting on the list and changes the list from blocking
 * to non-blocking.  This operation is one-way and cannot be reversed later.
 * This function does not call any POSIX cancellation points. */
void
exs_p_list_unblock(struct our_list_head *list)
{
	int wc;
	if ((list->flags & OUR_LIST_BLOCKING)) {
		pthread_mutex_lock(&list->tail_mutex);
		atomic_clear_bits(OUR_LIST_BLOCKING, &list->flags);
		wc = list->wait_count;
		pthread_mutex_unlock(&list->tail_mutex);
		if (wc) {
			pthread_cond_broadcast(&list->empty_cv);
		}
	}
}

/* Prior to calling this method, the caller must ensure through external means
 * that no more list manipulations will be done on l2.  List l2 will be
 * invalidated as part of the implementation of this function.  This operation
 * is not allowed if either list is blocking. This function does not call any
 * POSIX cancellation points. */
int
exs_p_list_splice(struct our_list_head *l1, struct our_list_head *l2)
{
	struct our_list_slot *old_l2_head;

	if ((l1->flags&OUR_LIST_BLOCKING) || (l2->flags&OUR_LIST_BLOCKING)) {
		errno = EINVAL;
		return -1;
	}

	pthread_mutex_lock(&l1->head_mutex);
	pthread_mutex_lock(&l1->tail_mutex);
	pthread_mutex_lock(&l2->head_mutex);
	pthread_mutex_lock(&l2->tail_mutex);

	old_l2_head = l2->head;
	l1->tail->next = l2->head->next;
	l1->tail = l2->tail;

	l1->empty_tail->next = l2->empty_head;
	l1->empty_tail = l2->empty_head;

	l2->tail = l2->head = NULL;
	l2->empty_tail = l2->empty_head = NULL;
	pthread_mutex_unlock(&l2->tail_mutex);
	pthread_mutex_unlock(&l2->head_mutex);
	pthread_mutex_destroy(&l2->head_mutex);
	pthread_mutex_destroy(&l2->tail_mutex);

	put_free_node(l1, old_l2_head);

	pthread_mutex_unlock(&l1->tail_mutex);
	pthread_mutex_unlock(&l1->head_mutex);

	return 0;
}
